#pragma once

/* template_opts.h
 * Additional template header file containing solver-specific model options.
 * Empty by default, intended to be written to by any solver to provide unique data.
 * 
 * Default options are still provided in the template_defaults.h header file,
 *   providing solver-specific defaults with the GPY_SOLVER_* macro.
 */
