"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# =============================================================================
# @file    __version__.py
# @brief   GillesPy2 version info
# @license Please see the file named LICENSE in the project directory
# @website https://github.com/StochSS/GillesPy2
# =============================================================================


__version__      = '1.6.6'

__title__        = 'GillesPy2'
__description__  = 'Python interface for Gillespie-style biochemical simulations'
__url__          = 'https://github.com/StochSS/GillesPy2'
__download_url__ = 'https://pypi.org/project/gillespy2/#files'
__author__       = 'See AUTHORS'
__email__        = 'bdrawert@unca.edu'
__license__      = 'GPL'
__copyright__    = 'Copyright (C) 2017-2021'
